#include "defines.h"
#include "serial.h"

#define IO_OFFSET 0x20

#define MEM8(addr) (*(volatile unsigned char *)(addr))
#define IO8(addr)  MEM8((addr) + IO_OFFSET)

#define USART_UCSR0A MEM8(0xc0)

#define USART_UCSR0A_UDRE0 (1<<5) /* Хåեǥå */
#define USART_UCSR0A_TXC0  (1<<6) /* λǥå */
#define USART_UCSR0A_RXC0  (1<<7) /* Хåե˥ǡͭǥå */

#define USART_UCSR0B MEM8(0xc1)
#define USART_UCSR0B_UCSZ02 (1<<2)
#define USART_UCSR0B_TXEN0  (1<<3)
#define USART_UCSR0B_RXEN0  (1<<4)
#define USART_UCSR0B_UDRIE0 (1<<5)
#define USART_UCSR0B_TXCIE0 (1<<6)
#define USART_UCSR0B_RXCIE0 (1<<7)

#define USART_UCSR0C MEM8(0xc2)
#define USART_UCSR0C_UCSZ00      (1<<1)
#define USART_UCSR0C_UCSZ01      (1<<2)
#define USART_UCSR0C_STOPBIT_1   (0<<3)
#define USART_UCSR0C_STOPBIT_2   (1<<3)
#define USART_UCSR0C_PARITY_NONE (0<<4)
#define USART_UCSR0C_PARITY_EVEN (2<<4)
#define USART_UCSR0C_PARITY_ODD  (3<<4)
#define USART_UCSR0C_MODE_ASYNC  (0<<6)
#define USART_UCSR0C_MODE_SYNC   (1<<6)
#define USART_UCSR0C_MODE_SPI    (3<<6)

#define USART_UBRR0L MEM8(0xc4)
#define USART_UBRR0H MEM8(0xc5)

#define USART_UDR0   MEM8(0xc6)

#define PIN_DDRD     IO8(0x0a)
#define PIN_PORTD    IO8(0x0b)
#define PIN_DDD0     (1<<0)
#define PIN_PORTD0   (1<<0)

#define MCUCR_IO    IO8(0x35)
#define MCUCR_IVCE  (1<<0)
#define MCUCR_IVSEL (1<<1)

/* ǥХ */
int serial_init(int index)
{
  MCUCR_IO = MCUCR_IVCE;
  MCUCR_IO = 0;

  USART_UCSR0B = 0;

#define CPU_CLOCK 16000000L
#define BAUDRATE 9600
  USART_UBRR0L = (CPU_CLOCK / (BAUDRATE * 16L) - 1);
  USART_UBRR0H = (CPU_CLOCK / (BAUDRATE * 16L) - 1) >> 8;

  USART_UCSR0B = USART_UCSR0B_RXEN0 | USART_UCSR0B_TXEN0; /* ǽ */
  USART_UCSR0C = USART_UCSR0C_UCSZ00 | USART_UCSR0C_UCSZ01;

  PIN_DDRD  &= ~PIN_DDD0;
  PIN_PORTD |=  PIN_PORTD0;

  return 0;
}

/* ǽ */
int serial_is_send_enable(int index)
{
  return (USART_UCSR0A & USART_UCSR0A_UDRE0);
}

/* ʸ */
int serial_send_byte(int index, unsigned char c)
{
  /* ǽˤʤޤԤ */
  while (!serial_is_send_enable(index))
    ;
  USART_UDR0 = c;
#ifdef SIMULATOR
  USART_UCSR0A &= ~USART_UCSR0A_TXC0; /*  */
#endif

  return 0;
}

/* ǽ */
int serial_is_recv_enable(int index)
{
  return (USART_UCSR0A & USART_UCSR0A_RXC0);
}

/* ʸ */
unsigned char serial_recv_byte(int index)
{
  unsigned char c;

  /* ʸޤԤ */
  while (!serial_is_recv_enable(index))
    ;
  c = USART_UDR0;
#ifdef SIMULATOR
  USART_UCSR0A &= ~USART_UCSR0A_RXC0; /* λ */
#endif

  return c;
}

/* ͭ */
int serial_intr_is_send_enable(int index)
{
  return (USART_UCSR0B & USART_UCSR0B_TXCIE0) ? 1 : 0;
}

/* ͭ */
void serial_intr_send_enable(int index)
{
  USART_UCSR0B |= USART_UCSR0B_TXCIE0;
}

/* ̵ */
void serial_intr_send_disable(int index)
{
  USART_UCSR0B &= ~USART_UCSR0B_TXCIE0;
}

/* ͭ */
int serial_intr_is_recv_enable(int index)
{
  return (USART_UCSR0B & USART_UCSR0B_UDRIE0) ? 1 : 0;
}

/* ͭ */
void serial_intr_recv_enable(int index)
{
  USART_UCSR0B |= USART_UCSR0B_UDRIE0;
}

/* ̵ */
void serial_intr_recv_disable(int index)
{
  USART_UCSR0B &= ~USART_UCSR0B_UDRIE0;
}
